# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2024 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from copy import deepcopy
from kubernator.api import jp
from jsonpath_ng.jsonpath import Index, Child


# https://github.com/kubernetes/community/blob/master/contributors/devel/sig-api-machinery/strategic-merge-patch.md

def extract_merge_instructions(manifest, resource):
    normalized_manifest = deepcopy(manifest)
    change_instrs = jp('$..`match(/\\\\$patch|\\\\$deleteFromPrimitiveList\\/.*/)`').find(manifest)

    instructions = []
    for change_instr in change_instrs:
        field = change_instr.path.fields[0]
        context = change_instr.context.value

        list_of_maps = False
        search_key = None
        change_path = change_instr.full_path.left
        if isinstance(change_path, Child) and isinstance(change_path.right, Index):
            list_of_maps = True
            index = change_path.right.index
            change_path = change_path.left
            clean_manifest_result: list = change_path.find(normalized_manifest)[0].value
            search_key = context.copy()
            del search_key[field]
            del clean_manifest_result[index]
        else:
            clean_manifest_result = change_path.find(normalized_manifest)[0]
            clean_manifest_result_context = clean_manifest_result.value
            del clean_manifest_result_context[field]

        instruction_value = context[field]
        if field == "$patch":
            if instruction_value in ("replace", "delete"):
                instructions.append(("patch", instruction_value,
                                     change_path, list_of_maps, search_key))
            else:
                raise ValueError("Invalid $patch instruction %r in resource %s at %s" %
                                 (instruction_value,
                                  resource,
                                  change_path))
        elif field.startswith("$deleteFromPrimitiveList/"):
            instructions.append(("delete-from-list", instruction_value, field[25:],
                                 change_path))

    return instructions, normalized_manifest


def apply_merge_instructions(merge_instrs, source_manifest, target_manifest, logger, resource):
    for merge_instr in merge_instrs:
        if merge_instr[0] == "patch":
            op, op_type, change_path, list_of_maps, search_key = merge_instr
        else:
            op, delete_list, field_name, change_path = merge_instr

        if op == "patch":
            source_obj = change_path.find(source_manifest)[0].value
            merged_obj = change_path.find(target_manifest)[0].value
            if op_type == "delete":
                if list_of_maps:
                    logger.trace("Deleting locally in resource %s: %s from %s at %s",
                                 resource,
                                 search_key,
                                 merged_obj,
                                 change_path)
                    del_idxs = []
                    for idx, obj in enumerate(merged_obj):
                        for k, v in search_key.items():
                            if k in obj:
                                if v is not None and obj[k] == v:
                                    del_idxs.append(idx)

                    for idx in del_idxs:
                        del merged_obj[idx]
                else:
                    logger.trace("Deleting locally in resource %s: %s at %s",
                                 resource,
                                 merged_obj,
                                 change_path)
                    merged_datum = change_path.find(target_manifest)[0]
                    merged_datum.context.value[merged_datum.path.fields[0]] = None
            elif op_type == "replace":
                logger.trace("Replacing locally in resource %s: %s with %s at %s",
                             resource,
                             merged_obj, source_obj,
                             change_path)
                merged_obj.clear()
                if list_of_maps:
                    merged_obj.extend(source_obj)
                else:
                    merged_obj.update(source_obj)
            else:
                raise ValueError("Invalid $patch instruction %s found at %s in resource %s",
                                 op_type, change_path, resource)

        elif op == "delete-from-list":
            merged_list: list = change_path.find(target_manifest)[0].value[field_name]
            if not isinstance(merged_list, list):
                raise ValueError("Not a list in resource %s: %s in %r at %s" %
                                 (resource, merged_list, field_name, change_path))
            logger.trace("Deleting from list locally in resource %s: %s from %s in %r at %s",
                         resource, delete_list, merged_list, field_name, change_path)
            for v in delete_list:
                try:
                    merged_list.remove(v)
                except ValueError:
                    logger.warning("No value %s to delete from list %s in %r at %s in resource %s",
                                   v, merged_list, field_name, change_path, resource)
        else:
            raise RuntimeError("should never reach here")



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2021 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import argparse
import datetime
import importlib
import logging
import pkgutil
import sys
import urllib.parse
from collections import deque
from collections.abc import MutableMapping, Callable
from pathlib import Path
from shutil import rmtree
from typing import Optional, Union

import yaml

import kubernator
from kubernator.api import (KubernatorPlugin, Globs, scan_dir, PropertyDict, config_as_dict, config_parent,
                            download_remote_file, load_remote_file, Repository, StripNL, jp, get_app_cache_dir,
                            get_cache_dir, install_python_k8s_client)
from kubernator.proc import run, run_capturing_out, run_pass_through_capturing

TRACE = 5


def trace(self, msg, *args, **kwargs):
    """
    Log 'msg % args' with severity 'TRACE'.

    To pass exception information, use the keyword argument exc_info with
    a true value, e.g.

    logger.trace("Houston, we have a %s", "interesting problem", exc_info=1)
    """
    if self.isEnabledFor(TRACE):
        self._log(TRACE, msg, args, **kwargs)


logging.addLevelName(5, "TRACE")
logging.Logger.trace = trace
logger = logging.getLogger("kubernator")

try:
    del (yaml.resolver.Resolver.yaml_implicit_resolvers["="])
except KeyError:
    pass


def define_arg_parse():
    parser = argparse.ArgumentParser(prog="kubernator",
                                     description="Kubernetes Provisioning Tool",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    g = parser.add_mutually_exclusive_group()
    g.add_argument("--version", action="version", version=kubernator.__version__,
                   help="print version and exit")
    g.add_argument("--clear-cache", action="store_true",
                   help="clear cache and exit")
    g.add_argument("--clear-k8s-cache", action="store_true",
                   help="clear Kubernetes Client cache and exit")
    g.add_argument("--pre-cache-k8s-client", action="extend", nargs="+", type=int,
                   help="download specified K8S client library major(!) version(s) and exit")
    parser.add_argument("--pre-cache-k8s-client-no-patch", action="store_true", default=None,
                        help="do not patch the k8s client being pre-cached")
    parser.add_argument("--log-format", choices=["human", "json"], default="human",
                        help="whether to log for human or machine consumption")
    parser.add_argument("--log-file", type=argparse.FileType("w"), default=None,
                        help="where to log, defaults to `stderr`")
    parser.add_argument("-v", "--verbose", choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE"],
                        default="INFO", help="how verbose do you want Kubernator to be")
    parser.add_argument("-f", "--file", type=argparse.FileType("w"), default=sys.stdout,
                        help="where to generate results, if necessary")
    parser.add_argument("-o", "--output-format", choices=["json", "json-pretty", "yaml"], default="yaml",
                        help="in what format to generate results")
    parser.add_argument("-p", "--path", dest="path", default=".",
                        type=Path, help="path to start processing")
    #    parser.add_argument("--pre-start-script", default=None, type=Path,
    #                        help="location of the pre-start script")
    #    parser.add_argument("--disconnected", action="store_true", default=False,
    #                        help="do not actually connect to the target Kubernetes")
    #    parser.add_argument("--k8s-version", type=str, default=None,
    #                        help="specify a version of Kubernetes when operating in the disconnected mode")
    parser.add_argument("--yes", action="store_false", default=True, dest="dry_run",
                        help="actually make destructive changes")
    parser.add_argument("command", nargs="?", choices=["dump", "apply"], default="dump",
                        help="whether to dump the proposed changes to the output or to apply them")
    return parser


def init_logging(verbose, output_format, output_file):
    root_log = logging.root

    handler = logging.StreamHandler(output_file)
    root_log.addHandler(handler)

    if output_format == "human":
        if handler.stream.isatty():
            import coloredlogs
            fmt_cls = coloredlogs.ColoredFormatter

        else:
            fmt_cls = logging.Formatter

        def formatTime(record, datefmt=None):
            return datetime.datetime.fromtimestamp(record.created).isoformat()

        formatter = fmt_cls("%(asctime)s %(name)s %(levelname)s %(filename)s:%(lineno)d %(message)s")
        formatter.formatTime = formatTime
    else:
        import json_log_formatter

        class JSONFormatter(json_log_formatter.JSONFormatter):
            def json_record(self, message, extra, record: logging.LogRecord):
                extra = super(JSONFormatter, self).json_record(message, extra, record)
                extra["ts"] = datetime.datetime.fromtimestamp(record.created)
                extra["name"] = record.name
                extra["level"] = record.levelname
                extra["fn"] = record.filename
                extra["ln"] = record.lineno
                del extra["time"]
                return extra

        formatter = JSONFormatter()

    handler.setFormatter(formatter)
    logger.setLevel(logging._nameToLevel[verbose])


class App(KubernatorPlugin):
    _name = "app"

    def __init__(self, args):
        self.args = args
        path = args.path.absolute()

        global_context = PropertyDict()
        global_context.globals = global_context
        context = PropertyDict(_parent=global_context)
        context._plugins = []

        self._top_level_context = context
        self.context = context
        self._top_dir_context = PropertyDict(_parent=self.context)

        self.repos: MutableMapping[Repository, Repository] = dict()
        self.path_q: deque[tuple[PropertyDict, Path]] = deque(((self._top_dir_context, path),))

        self._new_paths: list[tuple[PropertyDict, Path]] = []

        self._cleanups = []
        self._plugin_types = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def run(self):
        logger.info("Starting Kubernator version %s", kubernator.__version__)

        self.register_plugin(self)

        try:
            try:
                while True:
                    cwd = self.next()
                    if not cwd:
                        logger.debug("No paths left to traverse")
                        break

                    context = self.context

                    logger.debug("Inspecting directory %s", self._display_path(cwd))
                    self._run_handlers(KubernatorPlugin.handle_before_dir, False, context, None, cwd)

                    if (ktor_py := (cwd / ".kubernator.py")).exists():
                        self._run_handlers(KubernatorPlugin.handle_before_script, False, context, None, cwd)

                        for h in self.context._plugins:
                            h.set_context(context)

                        self._exec_ktor(ktor_py)

                        for h in self.context._plugins:
                            h.set_context(None)

                        self._run_handlers(KubernatorPlugin.handle_after_script, True, context, None, cwd)

                    self._run_handlers(KubernatorPlugin.handle_after_dir, True, context, None, cwd)

                self.context = self._top_dir_context
                context = self.context

                self._run_handlers(KubernatorPlugin.handle_apply, True, context, None)

                self._run_handlers(KubernatorPlugin.handle_verify, True, context, None)
            finally:
                self.context = self._top_dir_context
                context = self.context
                self._run_handlers(KubernatorPlugin.handle_shutdown, True, context, None)
        except:  # noqa E722
            raise
        else:
            self.context = self._top_dir_context
            context = self.context
            self._run_handlers(KubernatorPlugin.handle_summary, True, context, None)

    def discover_plugins(self):
        importlib.invalidate_caches()
        search_path = Path(kubernator.__path__[0], "plugins")
        [importlib.import_module(name)
         for finder, name, is_pkg in
         pkgutil.iter_modules([str(search_path)], "kubernator.plugins.")]

        for plugin in KubernatorPlugin.__subclasses__():
            if plugin._name in self._plugin_types:
                logger.warning("Plugin named %r in %r is already reserved by %r and will be ignored",
                               plugin._name,
                               plugin,
                               self._plugin_types[plugin._name])
            else:
                logger.info("Plugin %r discovered in %r", plugin._name, plugin)
                self._plugin_types[plugin._name] = plugin

    def register_plugin(self, plugin: Union[KubernatorPlugin, type, str], **kwargs):
        context = self.context
        if isinstance(plugin, str):
            try:
                plugin_obj = self._plugin_types[plugin]()
            except KeyError:
                logger.critical("No known plugin with the name %r", plugin)
                raise RuntimeError("No known plugin with the name %r" % (plugin,))
        elif isinstance(plugin, type):
            plugin_obj = plugin()
        else:
            plugin_obj = plugin

        for p in context._plugins:
            if p._name == plugin_obj._name:
                logger.info("Plugin with name %r already registered, skipping", p._name)
                return

        logger.info("Registering plugin %r via %r", plugin_obj._name, plugin_obj)

        # Register
        self._run_handlers(KubernatorPlugin.register, False, context, plugin_obj, **kwargs)

        context._plugins.append(plugin_obj)

        # Init
        self._run_handlers(KubernatorPlugin.handle_init, False, context, plugin_obj)

        # Start
        self._run_handlers(KubernatorPlugin.handle_start, False, context, plugin_obj)

        # If we're already processing a directory
        if "app" in self.context and "cwd" in self.context.app and self.context.app.cwd:
            cwd = self.context.app.cwd
            self._run_handlers(KubernatorPlugin.handle_before_dir, False, context, plugin_obj, cwd)

            # If we're already in the script (TODO: is it possible to NOT be in a script?)
            if "script" in self.context.app:
                self._run_handlers(KubernatorPlugin.handle_before_script, False, context, plugin_obj, cwd)

    def _run_handlers(self, __f, __reverse, __context, __plugin, *args, **kwargs):
        f_name = __f.__name__

        def run(h):
            h_f = getattr(h, f_name, None)
            if h_f:
                logger.trace("Running %r handler on %r with %r, %r", f_name, h, args, kwargs)
                h_f(*args, **kwargs)

        if __plugin:
            __plugin.set_context(__context)
            run(__plugin)
        else:
            self._set_plugin_context(__reverse, __context, run)

    def _set_plugin_context(self, reverse, context, run):
        for h in list(self.context._plugins if not reverse else reversed(self.context._plugins)):
            h.set_context(context)
            run(h)
            h.set_context(None)

    def _exec_ktor(self, ktor_py: Path):
        ktor_py_display_path = self._display_path(ktor_py)
        logger.debug("Executing %s", ktor_py_display_path)
        with open(ktor_py, "rb") as f:
            source = f.read()
        co = compile(source, ktor_py_display_path, "exec")
        globs = {"ktor": self.context,
                 "logger": logger.getChild("script")
                 }
        exec(co, globs)
        logger.debug("Executed %r", ktor_py_display_path)

    def next(self) -> Path:
        path_queue: deque[tuple[PropertyDict, Path]] = self.path_q
        if path_queue:
            self.context, path = path_queue.pop()
            return path

    def register_cleanup(self, h):
        if not hasattr(h, "cleanup"):
            raise RuntimeError("cleanup handler has no cleanup attribute")
        self._cleanups.append(h)

    def cleanup(self):
        for h in self._cleanups:
            h.cleanup()

    def register(self, **kwargs):
        self.discover_plugins()

    def handle_init(self):
        context = self.context

        context.globals.common = dict()
        context.globals.app = dict(display_path=self._display_path,
                                   args=self.args,
                                   repository_credentials_provider=self._repository_credentials_provider,
                                   walk_remote=self.walk_remote,
                                   walk_local=self.walk_local,
                                   register_plugin=self.register_plugin,
                                   config_as_dict=config_as_dict,
                                   config_parent=config_parent,
                                   download_remote_file=download_remote_file,
                                   load_remote_file=load_remote_file,
                                   register_cleanup=self.register_cleanup,
                                   jp=jp,
                                   run=self._run,
                                   run_capturing_out=self._run_capturing_out,
                                   run_passthrough_capturing=self._run_passthrough_capturing,
                                   repository=self.repository,
                                   StripNL=StripNL,
                                   default_includes=Globs(["*"], True),
                                   default_excludes=Globs([".*"], True),
                                   )
        context.app = dict(_repository_credentials_provider=None,
                           default_includes=Globs(context.app.default_includes),
                           default_excludes=Globs(context.app.default_excludes),
                           includes=Globs(context.app.default_includes),
                           excludes=Globs(context.app.default_excludes),
                           )

    def handle_before_dir(self, cwd: Path):
        context = self.context
        app = context.app
        app.includes = Globs(app.default_includes)
        app.excludes = Globs(app.default_excludes)
        app.cwd = cwd
        self._new_paths = []

    def handle_before_script(self, cwd: Path):
        context = self.context
        app = context.app
        app.script = (cwd / ".kubernator.py")

    def handle_after_script(self, cwd: Path):
        context = self.context
        app = context.app
        del app.script

    def handle_after_dir(self, cwd: Path):
        context = self.context
        app = context.app

        for f in scan_dir(logger, cwd, lambda d: d.is_dir(), app.excludes, app.includes):
            self._new_paths.append((PropertyDict(_parent=context), f))

        self.path_q.extend(reversed(self._new_paths))

        del app.cwd

    def repository(self, repo):
        repository = Repository(repo, self._repo_cred_augmentation)
        if repository in self.repos:
            repository = self.repos[repository]
        else:
            self.repos[repository] = repository
            repository.init(logger, self.context)
            self.register_cleanup(repository)

        return repository

    def walk_local(self, *paths: Union[Path, str, bytes], keep_context=False):
        for path in paths:
            p = Path(path)
            if not p.is_absolute():
                p = self.context.app.cwd / p
            self._add_local(p, keep_context)

    def walk_remote(self, repo, *path_prefixes: Union[Path, str, bytes], keep_context=False):
        repository = self.repository(repo)

        if path_prefixes:
            for path_prefix in path_prefixes:
                path = Path(path_prefix)
                if path.is_absolute():
                    path = Path(*path.parts[1:])
                self._add_local(repository.local_dir / path, keep_context)
        else:
            self._add_local(repository.local_dir, keep_context)

    def set_context(self, context):
        # We are managing the context for everyone so we don't actually set it anywhere
        pass

    def _add_local(self, path: Path, keep_context=False):
        logger.info("Adding %s to the plan%s", self._display_path(path),
                    " (in parent context)" if keep_context else "")
        self._new_paths.append((self.context if keep_context else PropertyDict(_parent=self.context), path))

    def _repository_credentials_provider(self,
                                         provider: Optional[
                                             Callable[[urllib.parse.SplitResult], tuple[
                                                 Optional[str], Optional[str], Optional[str]]]]):
        self.context.app._repository_credentials_provider = provider

    def _repo_cred_augmentation(self, url):
        rcp = self.context.app._repository_credentials_provider
        if not rcp:
            return url

        scheme, username, password = rcp(url)
        return urllib.parse.SplitResult(scheme if scheme else url.scheme,
                                        ((
                                             username +
                                             (
                                                 ":" + password if password else "") + "@"
                                             if username else "") + url.hostname)
                                        if username or password
                                        else url.netloc,
                                        url.path, url.query, url.fragment)

    def _path_to_repository(self, path: Path) -> Repository:
        for r in self.repos.values():
            if path.is_relative_to(r.local_dir):
                return r

    def _display_path(self, path: Path) -> str:
        repo = self._path_to_repository(path)
        return "<%s> %s" % (repo.url_str, path) if repo else str(path)

    def _run(self, *args, **kwargs):
        return run(*args, **kwargs)

    def _run_capturing_out(self, *args, **kwargs):
        return run_capturing_out(*args, **kwargs)

    def _run_passthrough_capturing(self, *args, **kwargs):
        return run_pass_through_capturing(*args, **kwargs)

    def __repr__(self):
        return "Kubernator"


def clear_cache():
    cache_dir = get_app_cache_dir()
    _clear_cache("Clearing application cache at %s", cache_dir)


def clear_k8s_cache():
    cache_dir = get_cache_dir("python")
    _clear_cache("Clearing Kubernetes Client cache at %s", cache_dir)


def _clear_cache(msg, cache_dir):
    logger.info(msg, cache_dir)
    if cache_dir.exists():
        rmtree(cache_dir)


def pre_cache_k8s_clients(*versions, disable_patching=False):
    proc_logger = logger.getChild("proc")
    stdout_logger = StripNL(proc_logger.info)
    stderr_logger = StripNL(proc_logger.warning)

    for v in versions:
        logger.info("Caching K8S client library ~=v%s.0%s...", v,
                    " (no patches)" if disable_patching else "")
        install_python_k8s_client(run_pass_through_capturing, v, logger, stdout_logger, stderr_logger, disable_patching)


def main():
    argparser = define_arg_parse()
    args = argparser.parse_args()
    if not args.pre_cache_k8s_client and args.pre_cache_k8s_client_no_patch is not None:
        argparser.error("--pre-cache-k8s-client-no-patch can only be used with --pre-cache-k8s-client")

    init_logging(args.verbose, args.log_format, args.log_file)

    try:
        if args.clear_cache:
            clear_cache()
            return

        if args.clear_k8s_cache:
            clear_k8s_cache()
            return

        if args.pre_cache_k8s_client:
            pre_cache_k8s_clients(*args.pre_cache_k8s_client,
                                  disable_patching=args.pre_cache_k8s_client_no_patch)
            return

        with App(args) as app:
            app.run()
    except SystemExit as e:
        return e.code
    except Exception as e:
        logger.fatal("Kubernator terminated with an error: %s", e, exc_info=e)
        return 1
    else:
        logger.info("Kubernator terminated successfully")
    finally:
        logging.shutdown()



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2021 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import logging
import os
from pathlib import Path
from tempfile import TemporaryDirectory

import yaml

from kubernator.api import (KubernatorPlugin, scan_dir,
                            FileType,
                            load_remote_file,
                            io_StringIO,
                            TemplateEngine,
                            StripNL,
                            Globs, load_file)
from kubernator.plugins.k8s_api import K8SResourcePluginMixin
from kubernator.proc import CalledProcessError

logger = logging.getLogger("kubernator.kops")
proc_logger = logger.getChild("kops")
stdout_logger = StripNL(proc_logger.info)
stderr_logger = StripNL(proc_logger.warning)

KOPS_CRDS = ["kops.k8s.io_clusters.yaml",
             "kops.k8s.io_instancegroups.yaml",
             "kops.k8s.io_keysets.yaml",
             "kops.k8s.io_sshcredentials.yaml"]

OBJECT_SCHEMA_VERSION = "1.20.6"


class KopsPlugin(KubernatorPlugin, K8SResourcePluginMixin):
    logger = logger

    _name = "kops"

    def __init__(self):
        self.context = None
        self.kops_stanza = ["kops"]
        self.version = None
        self.kops_dir = None
        self.kops_path = None

        self.template_engine = TemplateEngine(logger)
        super().__init__()

    def set_context(self, context):
        self.context = context

    def register(self, helm_version=None):
        raise RuntimeError("Currently disabled, please don't use")
        self.context.app.register_plugin("kubeconfig")
        self.context.app.register_plugin("kubectl")

    def handle_init(self):
        context = self.context

        # Temp dir
        self.kops_dir = TemporaryDirectory()
        context.app.register_cleanup(self.kops_dir)
        self.kops_path = kops_path = Path(self.kops_dir.name)
        kops_config = kops_path / ".kops.yaml"
        kops_config.touch()
        self.kops_stanza.extend(("--config", str(kops_config)))

        # Kube config dir
        kubeconfig_dir = kops_path / ".kube"
        kubeconfig_dir.mkdir()
        self.kubeconfig_path = kubeconfig_dir / "config"

        # if logger.getEffectiveLevel() < logging.DEBUG:
        #    self.kops_stanza.append("-v5")
        # elif logger.getEffectiveLevel() < logging.INFO:
        #    self.kops_stanza.append("-v2")

        context.globals.kops = dict(walk_remote=self.walk_remote,
                                    walk_local=self.walk_local,
                                    update=self.update,
                                    export=self.export,
                                    master_interval="8m",
                                    node_interval="8m"
                                    )
        context.kops = dict()

        self.version = context.app.run_capturing_out(["kops", "version", "--short"], stderr_logger).strip()
        logger.info("Found kOps version %s", self.version)

        self.resource_definitions_schema = load_remote_file(logger,
                                                            f"https://raw.githubusercontent.com/kubernetes/kubernetes/"
                                                            f"v{OBJECT_SCHEMA_VERSION}/api/openapi-spec/swagger.json",
                                                            FileType.JSON)
        self._populate_resource_definitions()

        common_url_path = f"https://raw.githubusercontent.com/kubernetes/kops/v{self.version}/k8s/crds"
        for kops_crd in KOPS_CRDS:
            url = f"{common_url_path}/{kops_crd}"
            self.add_remote_crds(url, FileType.YAML, sub_category="kops")

    def handle_start(self):
        context = self.context

        # Exclude Kops YAMLs from K8S resource loading
        context.k8s.default_excludes.add("*.kops.yaml")
        context.k8s.default_excludes.add("*.kops.yml")

    def walk_remote(self, url, *paths: Path,
                    excludes=(".*",), includes=("*.kops.yaml", "*.kops.yml")):
        repository = self.context.app.repository(url)
        for path in paths:
            self.walk_local(repository.local_dir / path, excludes, includes)

    def walk_local(self, path: Path, excludes=(".*",), includes=("*.kops.yaml", "*.kops.yml")):
        context = self.context
        path = Path(path)

        for f in scan_dir(logger, path, lambda d: d.is_file(), Globs(excludes), Globs(includes)):
            p = path / f.name
            display_p = context.app.display_path(p)
            logger.debug("Adding Kops resources from %s", display_p)

            manifests = load_file(logger, p, FileType.YAML, display_p,
                                  self.template_engine,
                                  {"ktor": context})

            self.add_resources(manifests, display_p)

    def update(self):
        context = self.context
        run = context.app.run
        run_capturing_out = context.app.run_capturing_out

        os.environ["KUBECONFIG"] = str(self.kubeconfig_path)
        os.environ["KOPS_CLUSTER_NAME"] = context.kops.cluster_name
        os.environ["KOPS_STATE_STORE"] = context.kops.state_store

        kops_extra_args = ["--name", context.kops.cluster_name, "--state", context.kops.state_store]

        cmd = context.app.args.command
        dry_run = context.app.args.dry_run

        for resource in self.resources.values():
            logger.info("Replacing/creating kOps resource %s", resource)
            resource_out = io_StringIO()
            yaml.dump(resource.manifest, resource_out)
            if cmd != "apply" or dry_run:
                logger.info("Would replace kOps resource if not for dry-run mode: %s", resource_out.getvalue())
            else:
                run(self.kops_stanza + ["replace", "--force", "-f", "-"] + kops_extra_args,
                    stdout_logger,
                    stderr_logger,
                    resource_out.getvalue()).wait()

        logger.info("Staging kOps update")
        update_cmd = self.kops_stanza + ["update", "cluster"] + kops_extra_args
        result = run_capturing_out(update_cmd, stderr_logger)
        proc_logger.info(result)
        if "Must specify --yes to apply changes" in result:
            logger.info("kOps update would make changes")
            if cmd != "apply" or dry_run:
                logger.info("Skipping actual kOps update due to dry-run mode")
            else:
                logger.info("Running kOps update")
                run(update_cmd + ["--yes"],
                    stdout_logger,
                    stderr_logger).wait()

        self.export()

        if cmd != "apply":
            logger.info("Skipping cluster validation since not in apply mode")
        else:
            validation_failed = False
            try:
                output = run_capturing_out(self.kops_stanza + ["validate", "cluster", "-o", "json"],
                                           stderr_logger)
            except CalledProcessError as e:
                validation_failed = True
                output = e.output

            if validation_failed:
                if output:
                    validation_results = json.loads(output)
                    for failure in validation_results["failures"]:
                        logger.error("%s %s failed validation: %s", failure["type"], failure["name"],
                                     failure["message"])
                raise RuntimeError("Cluster validation failed!")
            else:
                logger.info("Cluster validation successful!")

        logger.info("Staging kOps cluster rolling update")
        rolling_update_cmd = self.kops_stanza + ["rolling-update", "cluster",
                                                 "--master-interval", context.kops.master_interval,
                                                 "--node-interval", context.kops.node_interval] + kops_extra_args
        result = run_capturing_out(rolling_update_cmd, stderr_logger)
        proc_logger.info(result)
        if "Must specify --yes to rolling-update" in result:
            logger.info("kOps cluster rolling update would make changes")
            if cmd != "apply" or dry_run:
                logger.info("Skipping actual kOps cluster rolling update due to dry-run")
            else:
                logger.info("Running kOps cluster rolling update")
                run(rolling_update_cmd + ["--yes"],
                    stdout_logger,
                    stderr_logger).wait()

    def export(self):
        context = self.context
        run = context.app.run

        kops_extra_args = ["--name", context.kops.cluster_name, "--state", context.kops.state_store]
        logger.info("Exporting kubeconfig from kOps")
        run(self.kops_stanza + ["export", "kubecfg", "--admin"] + kops_extra_args,
            stdout_logger,
            stderr_logger).wait()

        if not self.kubeconfig_path.exists():
            raise RuntimeError("kOps failed to export kubeconfig for unknown reason - check AWS credentials")

    def __repr__(self):
        return "kOps Plugin"



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2021 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import logging
import os
import tarfile
import tempfile
from pathlib import Path
from shutil import which

import yaml

from kubernator.api import (KubernatorPlugin, scan_dir,
                            TemplateEngine,
                            load_remote_file,
                            FileType,
                            StripNL,
                            Globs,
                            get_golang_os,
                            get_golang_machine,
                            prepend_os_path, jp, load_file)
from kubernator.plugins.k8s import api_exc_normalize_body, api_exc_format_body
from kubernator.plugins.k8s_api import K8SResourcePluginMixin

logger = logging.getLogger("kubernator.istio")
proc_logger = logger.getChild("proc")
stdout_logger = StripNL(proc_logger.info)
stderr_logger = StripNL(proc_logger.warning)

MESH_PILOT_JP = jp('$.meshVersion[?Component="pilot"].Info.version')


class IstioPlugin(KubernatorPlugin, K8SResourcePluginMixin):
    logger = logger
    _name = "istio"

    def __init__(self):
        self.context = None
        self.client_version = None
        self.server_version = None
        self.provision_operator = False
        self.install = False
        self.upgrade = False
        self.upgrade_from_operator = False
        self.template_engine = TemplateEngine(logger)

        self.istioctl_dir = None

        super().__init__()

    def register(self, version=None):
        context = self.context
        context.app.register_plugin("kubeconfig")
        context.app.register_plugin("k8s")

        if version:
            # Download and use specific version
            istioctl_os = get_golang_os()
            if istioctl_os == "darwin":
                istioctl_os = "osx"
                istioctl_machine = get_golang_machine()
                if istioctl_machine == "amd64":
                    istioctl_platform = istioctl_os
                else:
                    istioctl_platform = f"{istioctl_os}-{istioctl_machine}"
            else:
                istioctl_platform = f"{istioctl_os}-{get_golang_machine()}"
            istioctl_url = (f"https://github.com/istio/istio/releases/download/{version}/"
                            f"istioctl-{version}-{istioctl_platform}.tar.gz")
            istioctl_file_dl, _ = context.app.download_remote_file(logger, istioctl_url, "bin")
            istioctl_file_dl = str(istioctl_file_dl)
            self.istioctl_dir = tempfile.TemporaryDirectory()
            context.app.register_cleanup(self.istioctl_dir)

            istioctl_file = str(Path(self.istioctl_dir.name) / "istioctl")
            istio_tar = tarfile.open(istioctl_file_dl)
            istio_tar.extractall(self.istioctl_dir.name)

            os.chmod(istioctl_file, 0o500)
            prepend_os_path(self.istioctl_dir.name)
        else:
            # Use current version
            istioctl_file = which("istioctl")
            if not istioctl_file:
                raise RuntimeError("`istioctl` cannot be found and no version has been specified")

            logger.debug("Found istioctl in %r", istioctl_file)

        context.globals.istio = dict(
            default_includes=Globs(["*.istio.yaml", "*.istio.yml"], True),
            default_excludes=Globs([".*"], True),
            istioctl_file=istioctl_file,
            stanza=self.stanza,
            test=self.test_istioctl
        )

    def test_istioctl(self):
        context = self.context
        version_out: str = context.app.run_capturing_out(context.istio.stanza() + ["version", "-o", "json"],
                                                         stderr_logger)

        version_out_js = json.loads(version_out)
        version = version_out_js["clientVersion"]["version"]
        logger.info("Using istioctl %r version %r with stanza %r",
                    self.context.istio.istioctl_file, version, context.istio.stanza())

        logger.info("Found Istio client version %s", version)

        return version, version_out_js

    def set_context(self, context):
        self.context = context

    def stanza(self):
        context = self.context.istio
        return [context.istioctl_file, f"--kubeconfig={self.context.kubeconfig.kubeconfig}"]

    def handle_init(self):
        pass

    def handle_start(self):
        context = self.context

        version, version_out_js = self.test_istioctl()
        self.client_version = tuple(map(int, version.split(".")))
        mesh_versions = set(tuple(map(int, m.value.split("."))) for m in MESH_PILOT_JP.find(version_out_js))

        if mesh_versions:
            self.server_version = max(mesh_versions)

        if not self.server_version:
            logger.info("No Istio mesh has been found and it'll be created")
            self.install = True
            self.provision_operator = True
        elif self.server_version != self.client_version:
            logger.info("Istio client is version %s while server is up to %s - up/downgrade will be performed",
                        ".".join(map(str, self.client_version)),
                        ".".join(map(str, self.server_version)))
            self.upgrade = True
            self.provision_operator = True

        if self.client_version >= (1, 24, 0):
            # No more operator in 1.24.0+
            self.provision_operator = False

        if self.upgrade and (self.client_version >= (1, 24, 0) > self.server_version):
            self.upgrade_from_operator = True

        if self.upgrade and (self.client_version < (1, 24, 0) <= self.server_version):
            raise ValueError(f"Unable to downgrade Istio from {self.server_version} to {self.client_version}")

        # Register Istio-related CRDs with K8S
        if self.client_version >= (1, 24, 0):
            crd_path = "manifests/charts/base/files/crd-all.gen.yaml"
        else:
            crd_path = "manifests/charts/base/crds/crd-all.gen.yaml"
        self.context.k8s.load_remote_crds(
            f"https://raw.githubusercontent.com/istio/istio/{'.'.join(map(str, self.client_version))}/{crd_path}",
            "yaml")

        # This plugin only deals with Istio Operator, so only load that stuff
        self.resource_definitions_schema = load_remote_file(logger,
                                                            f"https://raw.githubusercontent.com/kubernetes/kubernetes/"
                                                            f"{self.context.k8s.server_git_version}"
                                                            f"/api/openapi-spec/swagger.json",
                                                            FileType.JSON)
        self._populate_resource_definitions()

        crd_operator_version = (1, 23, 4) if self.client_version >= (1, 24, 0) else self.client_version
        self.add_remote_crds(
            f"https://raw.githubusercontent.com/istio/istio/{'.'.join(map(str, crd_operator_version))}/"
            f"manifests/charts/istio-operator/crds/crd-operator.yaml", FileType.YAML)

        # Exclude Istio YAMLs from K8S resource loading
        context.k8s.default_excludes.add("*.istio.yaml")
        context.k8s.default_excludes.add("*.istio.yml")

    def handle_before_dir(self, cwd: Path):
        context = self.context

        context.istio.default_includes = Globs(context.istio.default_includes)
        context.istio.default_excludes = Globs(context.istio.default_excludes)
        context.istio.includes = Globs(context.istio.default_includes)
        context.istio.excludes = Globs(context.istio.default_excludes)

        # Exclude Istio YAMLs from K8S resource loading
        context.k8s.excludes.add("*.istio.yaml")
        context.k8s.excludes.add("*.istio.yml")

    def handle_after_dir(self, cwd: Path):
        context = self.context
        istio = context.istio

        for f in scan_dir(logger, cwd, lambda d: d.is_file(), istio.excludes, istio.includes):
            p = cwd / f.name
            display_p = context.app.display_path(p)
            logger.info("Adding Istio Operator from %s", display_p)

            manifests = load_file(logger, p, FileType.YAML, display_p,
                                  self.template_engine,
                                  {"ktor": context})

            self.add_resources(manifests, display_p)

    def handle_apply(self):
        context = self.context

        if not self.resources:
            logger.info("Skipping Istio as no Operator was processed")
        else:
            with tempfile.NamedTemporaryFile(mode="wt", delete=False) as operators_file:
                logger.info("Saving Istio Operators to %s", operators_file.name)
                yaml.safe_dump_all((r.manifest for r in self.resources.values()), operators_file)

            if context.app.args.command == "apply":
                logger.info("Running Istio precheck")
                context.app.run(context.istio.stanza() + ["x", "precheck"],
                                stdout_logger, stderr_logger).wait()
                context.app.run(context.istio.stanza() + ["validate", "-f", operators_file.name],
                                stdout_logger, stderr_logger).wait()

                dry_run = context.app.args.dry_run

                if self.provision_operator:
                    self._create_istio_system_ns(True)
                    self._operator_init(operators_file, True)

                    if not dry_run:
                        self._create_istio_system_ns(False)
                        self._operator_init(operators_file, False)
                elif self.install:
                    self._install(operators_file, True)

                    if not dry_run:
                        self._install(operators_file, False)
                elif self.upgrade:
                    def _upgrade(dry_run):
                        if self.upgrade_from_operator:
                            # delete deployment -n istio-system istio-operator
                            self._delete_resource_internal({"apiVersion": "apps/v1",
                                                            "kind": "Deployment",
                                                            "metadata":
                                                                {
                                                                    "namespace": "istio-system",
                                                                    "name": "istio-operator"
                                                                }
                                                            }, dry_run, True)
                        self._upgrade(operators_file, dry_run)

                    _upgrade(True)
                    if not dry_run:
                        _upgrade(False)

    def _delete_resource_internal(self, manifest, dry_run=True, missing_ok=False):
        from kubernetes import client
        from kubernetes.client.rest import ApiException

        context = self.context
        k8s_client = context.k8s.client

        res = self._create_resource(manifest)
        res.rdef.populate_api(client, k8s_client)
        try:
            res.delete(dry_run=dry_run)
        except ApiException as e:
            api_exc_normalize_body(e)
            try:
                skip = False
                if e.status == 404 and missing_ok:
                    skip = True
                if not skip:
                    raise
            except ApiException as e:
                api_exc_format_body(e)
                raise

        return res

    def _create_resource_internal(self, manifest, dry_run=True, exists_ok=False):
        from kubernetes import client
        from kubernetes.client.rest import ApiException

        context = self.context
        k8s_client = context.k8s.client

        res = self._create_resource(manifest)
        res.rdef.populate_api(client, k8s_client)
        try:
            res.create(dry_run=dry_run)
        except ApiException as e:
            skip = False
            api_exc_normalize_body(e)
            try:
                if e.status == 409:
                    status = e.body
                    if status["reason"] == "AlreadyExists" and exists_ok:
                        skip = True
                if not skip:
                    raise
            except ApiException as e:
                api_exc_format_body(e)
                raise

        return res

    def _install(self, operators_file, dry_run):
        context = self.context
        status_details = " (dry run)" if dry_run else ""

        logger.info("Running Istio install%s", status_details)
        istio_install_cmd = context.istio.stanza() + ["install", "-f", operators_file.name, "-y", "--verify"]
        context.app.run(istio_install_cmd + (["--dry-run"] if dry_run else []),
                        stdout_logger,
                        stderr_logger).wait()

    def _upgrade(self, operators_file, dry_run):
        context = self.context
        status_details = " (dry run)" if dry_run else ""

        logger.info("Running Istio upgrade%s", status_details)
        istio_upgrade_cmd = context.istio.stanza() + ["upgrade", "-f", operators_file.name, "-y", "--verify"]
        context.app.run(istio_upgrade_cmd + (["--dry-run"] if dry_run else []),
                        stdout_logger,
                        stderr_logger).wait()

    def _create_istio_system_ns(self, dry_run):
        status_details = " (dry run)" if dry_run else ""
        logger.info("Creating istio-system namespace%s", status_details)
        self._create_resource_internal({"apiVersion": "v1",
                                        "kind": "Namespace",
                                        "metadata": {
                                            "labels": {
                                                "istio-injection": "disabled"
                                            },
                                            "name": "istio-system"
                                        }
                                        },
                                       dry_run=dry_run,
                                       exists_ok=True
                                       )

    def _operator_init(self, operators_file, dry_run):
        context = self.context
        status_details = " (dry run)" if dry_run else ""

        logger.info("Running Istio operator init%s", status_details)
        istio_operator_init = context.istio.stanza() + ["operator", "init", "-f", operators_file.name]
        context.app.run(istio_operator_init + (["--dry-run"] if dry_run else []),
                        stdout_logger,
                        stderr_logger).wait()

    def __repr__(self):
        return "Istio Plugin"



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2021 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import logging
from collections.abc import Mapping
from pathlib import Path

from jsonschema import Draft7Validator

from kubernator.api import (KubernatorPlugin, Globs, scan_dir, load_file, FileType, calling_frame_source,
                            validator_with_defaults, TemplateEngine, Template)

logger = logging.getLogger("kubernator.template")

TEMPLATE_SCHEMA = {
    "definitions": {
        "define": {
            "properties": {
                "define": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string"
                            },
                            "defaults": {
                                "additionalProperties": True,
                                "type": "object",
                                "default": {},
                                "description": "names and values of the default properties"
                            },
                            "path": {
                                "type": "string",
                                "description": "path to a template file"
                            },
                        },
                        "required": ["name", "path"]
                    }
                }
            }
        },
        "apply": {
            "properties": {
                "apply": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string"
                            },
                            "values": {
                                "additionalProperties": True,
                                "type": "object",
                                "default": {},
                                "description": "names and values of the rendering properties"
                            },
                        },
                        "required": ["name"]
                    }
                }
            }
        }
    },
    "allOf": [
        {"$ref": "#/definitions/define"},
        {"$ref": "#/definitions/apply"}
    ]
}

Draft7Validator.check_schema(TEMPLATE_SCHEMA)
TEMPLATE_VALIDATOR_CLS: type[Draft7Validator] = validator_with_defaults(Draft7Validator)
TEMPLATE_VALIDATOR: Draft7Validator = TEMPLATE_VALIDATOR_CLS(TEMPLATE_SCHEMA, format_checker=Draft7Validator.FORMAT_CHECKER)


class TemplatePlugin(KubernatorPlugin):
    logger = logger

    _name = "templates"

    def __init__(self):
        self.context = None

        self.templates: dict[str, Template] = {}
        self.template_engine = TemplateEngine(logger)

    def set_context(self, context):
        self.context = context

    def handle_init(self):
        context = self.context
        context.globals.templates = dict(default_includes=Globs(["*.tmpl.yaml", "*.tmpl.yml"], True),
                                         default_excludes=Globs([".*"], True),
                                         render_template=self.render_template,
                                         apply_template=self.apply_template
                                         )

    def handle_before_dir(self, cwd: Path):
        context = self.context

        context.templates.default_includes = Globs(context.templates.default_includes)
        context.templates.default_excludes = Globs(context.templates.default_excludes)
        context.templates.includes = Globs(context.templates.default_includes)
        context.templates.excludes = Globs(context.templates.default_excludes)

        # Exclude Template YAMLs from K8S resource loading
        context.k8s.excludes.add("*.tmpl.yaml")
        context.k8s.excludes.add("*.tmpl.yml")

    def handle_after_dir(self, cwd: Path):
        context = self.context
        templates = context.templates

        for f in scan_dir(logger, cwd, lambda d: d.is_file(), templates.excludes, templates.includes):
            p = cwd / f.name
            display_p = context.app.display_path(p)
            logger.debug("Adding Kubernator template from %s", display_p)

            template_docs = load_file(logger, p, FileType.YAML, display_p)

            for template_doc in template_docs:
                self._process_template_doc(template_doc, display_p)

    def add_local_template(self, path):
        display_p = self.context.app.display_path(path)

        templates = load_file(logger, path, FileType.YAML, display_p)

        self.add_template(templates, display_p)

    def add_template(self, template_docs, source=None):
        if source is None:
            source = calling_frame_source()

        if isinstance(template_docs, Mapping):
            self._process_template_doc(template_docs, source)
        else:
            for template in template_docs:
                self._process_template_doc(template, source)

    def render_template(self, name, source, values=()) -> str:
        if name not in self.templates:
            raise ValueError(f"Template with a name {name} does not exist")

        template = self.templates[name]
        templ_source = _get_source(name, template.source, source)
        logger.debug("Rendering %s", templ_source)
        rendered_template = template.render(self.context, values)

        if self.template_engine.failures():
            raise ValueError(f"Unable to render template {templ_source} due to undefined required variables")

        return rendered_template

    def apply_template(self, name, values=(), source=None):
        if source is None:
            source = calling_frame_source()

        rendered_template = self.render_template(name, source, values)
        template = self.templates[name]
        templ_source = _get_source(name, template.source, source)
        logger.info("Applying %s", templ_source)
        self.context.k8s.add_resources(rendered_template, source=templ_source)

    def _process_template_doc(self, template_doc, source):
        logger.info("Processing Kubernator template from %s", source)
        errors = list(TEMPLATE_VALIDATOR.iter_errors(template_doc))
        if errors:
            for error in errors:
                self.logger.error("Error detected in Kubernator template from %s", source, exc_info=error)
            raise errors[0]

        for template in template_doc.get("define", ()):
            self._add_template(template, source)

        for apply in template_doc.get("apply", ()):
            self.apply_template(apply["name"], apply["values"], source)

    def _add_template(self, template, source):
        self._add_parsed_template(source, **{k.replace("-", "_"): v for k, v in template.items()})

    def _add_parsed_template(self, source, *, name, path, defaults):
        if name in self.templates:
            raise ValueError(f"Template with a name {name} already exists and cannot be overwritten")

        file = Path(path)
        if not file.is_absolute():
            file = self.context.app.cwd / file

        with open(file, "rt") as f:
            template = self.template_engine.from_string(f.read())

        self.templates[name] = Template(name, template, defaults, source=source, path=file)

    def __repr__(self):
        return "Template Plugin"


def _get_source(name, template_def_source, template_appl_source):
    return "template %r from %s%s" % (name, template_def_source,
                                      "" if template_def_source == template_appl_source else
                                      " in %s" % template_appl_source)



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2023 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import logging
import tempfile
from pathlib import Path

from kubernator.api import (KubernatorPlugin,
                            StripNL
                            )

logger = logging.getLogger("kubernator.eks")
proc_logger = logger.getChild("proc")
stdout_logger = StripNL(proc_logger.info)
stderr_logger = StripNL(proc_logger.warning)


class EksPlugin(KubernatorPlugin):
    logger = logger

    _name = "eks"

    def __init__(self):
        self.context = None
        self.kubeconfig_dir = tempfile.TemporaryDirectory()
        self.name = None
        self.region = None

        super().__init__()

    def set_context(self, context):
        self.context = context

    def register(self, *, name=None, region=None):
        context = self.context

        if not name:
            raise ValueError("`name` is required")

        self.name = name
        self.region = region

        context.app.register_plugin("kubeconfig")
        context.app.register_plugin("awscli")

        context.globals.eks = dict(kubeconfig=str(Path(self.kubeconfig_dir.name) / "config")
                                   )

    def handle_init(self):
        context = self.context

        self.context.app.run(context.awscli.stanza(
            "eks", "update-kubeconfig", "--name", self.name, "--kubeconfig", context.eks.kubeconfig,
            region=self.region),
            stdout_logger,
            stderr_logger).wait()

        context.kubeconfig.set(context.eks.kubeconfig)

    def handle_shutdown(self):
        self.kubeconfig_dir.cleanup()

    def __repr__(self):
        return "EKS Plugin"



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2023 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import logging
import os

from kubernator.api import (KubernatorPlugin,
                            StripNL
                            )

logger = logging.getLogger("kubernator.kubeconfig")
proc_logger = logger.getChild("proc")
stdout_logger = StripNL(proc_logger.info)
stderr_logger = StripNL(proc_logger.warning)


class KubeConfigPlugin(KubernatorPlugin):
    logger = logger

    _name = "kubeconfig"

    def __init__(self):
        self.context = None
        super().__init__()

    def set_context(self, context):
        self.context = context

    def register(self,
                 kubeconfig=None):
        context = self.context

        kubeconfig = kubeconfig or os.environ.get("KUBECONFIG", os.path.expanduser("~/.kube/config"))

        context.globals.kubeconfig = dict(set=self.set_kubeconfig,
                                          register_change_notifier=self.register_change_notifier,
                                          kubeconfig=kubeconfig,
                                          _notifiers=[]
                                          )

        self.set_kubeconfig(kubeconfig)

    def set_kubeconfig(self, kubeconfig):
        context = self.context
        context.kubeconfig.kubeconfig = kubeconfig

        for n in context.kubeconfig._notifiers:
            n()

    def register_change_notifier(self, notifier):
        context = self.context

        context.kubeconfig._notifiers.append(notifier)

    def __repr__(self):
        return "Kubeconfig Plugin"



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2023 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import io
import json
import logging
import os
import tempfile
from pathlib import Path
from shutil import which, copy

import yaml

from kubernator.api import (KubernatorPlugin,
                            prepend_os_path,
                            StripNL,
                            get_golang_os,
                            get_golang_machine
                            )

logger = logging.getLogger("kubernator.kubectl")
proc_logger = logger.getChild("proc")
stdout_logger = StripNL(proc_logger.info)
stderr_logger = StripNL(proc_logger.warning)


class KubectlPlugin(KubernatorPlugin):
    logger = logger

    _name = "kubectl"

    def __init__(self):
        self.context = None
        self.kubectl_dir = None
        super().__init__()

    def set_context(self, context):
        self.context = context

    def stanza(self):
        context = self.context.kubectl
        return [context.kubectl_file, f"--kubeconfig={self.context.kubeconfig.kubeconfig}"]

    def register(self,
                 version=None,
                 kubeconfig=None):
        context = self.context

        context.app.register_plugin("kubeconfig")

        if not version and "k8s" in context:
            version = ".".join(context.k8s.server_version)
            logger.info("Kubernetes plugin is running with server version %s - will use it", version)

        if version:
            # Download and use specific version
            kubectl_url = f"https://dl.k8s.io/release/v{version}/bin/" \
                          f"{get_golang_os()}/" \
                          f"{get_golang_machine()}/kubectl"
            kubectl_file_dl, _ = context.app.download_remote_file(logger, kubectl_url, "bin")
            kubectl_file_dl = str(kubectl_file_dl)
            self.kubectl_dir = tempfile.TemporaryDirectory()
            context.app.register_cleanup(self.kubectl_dir)

            kubectl_file = str(Path(self.kubectl_dir.name) / "kubectl")
            copy(kubectl_file_dl, kubectl_file)
            os.chmod(kubectl_file, 0o500)
            prepend_os_path(self.kubectl_dir.name)
        else:
            # Use current version
            kubectl_file = which("kubectl")
            if not kubectl_file:
                raise RuntimeError("`kubectl` cannot be found and no version has been specified")

            logger.debug("Found kubectl in %r", kubectl_file)

        context.globals.kubectl = dict(version=version,
                                       kubectl_file=kubectl_file,
                                       stanza=self.stanza,
                                       test=self.test_kubectl,
                                       run=self.run,
                                       run_capturing=self.run_capturing,
                                       get=self.get,
                                       )

        context.globals.kubectl.version = context.kubectl.test()

    def run_capturing(self, *args, **kwargs):
        return self.context.app.run_capturing_out(self.stanza() +
                                                  list(args),
                                                  stderr_logger,
                                                  **kwargs
                                                  )

    def run(self, *args, **kwargs):
        self.context.app.run(self.stanza() +
                             list(args),
                             stdout_logger,
                             stderr_logger,
                             **kwargs
                             ).wait()

    def get(self, resource_type, resource_name, namespace=None):
        args = ["get", resource_type, resource_name]
        if namespace:
            args += ["-n", namespace]
        args += ["-o", "yaml"]

        res = list(yaml.safe_load_all(io.StringIO(self.context.kubectl.run_capturing(*args))))
        if len(res):
            if len(res) > 1:
                return res
            return res[0]
        return None

    def test_kubectl(self):
        version_out: str = self.run_capturing("version", "--client=true", "-o", "json")

        version_out_js = json.loads(version_out)
        kubectl_version = version_out_js["clientVersion"]["gitVersion"][1:]

        logger.info("Using kubectl %r version %r with stanza %r",
                    self.context.kubectl.kubectl_file, kubectl_version, self.stanza())

        return kubectl_version

    def __repr__(self):
        return "Kubectl Plugin"



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2024 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import re
from functools import cache

from jsonpath_ng import JSONPath, DatumInContext
from jsonpath_ng.ext import parse as jp_parse, parser
from jsonpath_ng.ext.string import DefintionInvalid

__all__ = ["jp", "JPath"]


class JPath:
    def __init__(self, pattern):
        self.pattern = jp_parse(pattern)

    def find(self, val):
        return self.pattern.find(val)

    def all(self, val):
        return list(map(lambda x: x.value, self.find(val)))

    def first(self, val):
        """Returns the first element or None if it doesn't exist"""
        try:
            return next(map(lambda x: x.value, self.find(val)))
        except StopIteration:
            return None

    def only(self, val):
        """Returns the first and only element.
        Raises ValueError if more than one value found
        Raises KeyError if no value found
        """
        m = map(lambda x: x.value, self.find(val))
        try:
            v = next(m)
        except StopIteration:
            raise KeyError("no value found")
        try:
            next(m)
            raise ValueError("more than one value returned")
        except StopIteration:
            return v


@cache
def jp(pattern) -> JPath:
    return JPath(pattern)


MATCH = re.compile(r"match\(/(.*)(?<!\\)/\)")


class Match(JSONPath):
    """Direct node regex matcher

    Concrete syntax is '`match(/regex/)`'
    """

    def __init__(self, method=None):
        m = MATCH.match(method)
        if m is None:
            raise DefintionInvalid("%s is not valid" % method)
        self.expr = m.group(1).strip()
        self.regex = re.compile(self.expr)
        self.method = method

    def find(self, datum):
        datum = DatumInContext.wrap(datum)

        if hasattr(datum.path, "fields") and self.regex.match(datum.path.fields[0]):
            return [datum]
        return []

    def __eq__(self, other):
        return isinstance(other, Match) and self.method == other.method

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.method)

    def __str__(self):
        return '`match(/%s/)`' % (self.expr,)


old_p_jsonpath_named_operator = parser.ExtentedJsonPathParser.p_jsonpath_named_operator


def p_jsonpath_named_operator(self, p):
    "jsonpath : NAMED_OPERATOR"
    if p[1].startswith("match("):
        p[0] = Match(p[1])
    else:
        old_p_jsonpath_named_operator(self, p)


parser.ExtentedJsonPathParser.p_jsonpath_named_operator = p_jsonpath_named_operator



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2023 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import logging
import tempfile
import os
from pathlib import Path
from shutil import which, copy

from kubernator.api import (KubernatorPlugin, Globs, StripNL,
                            scan_dir,
                            get_golang_os,
                            get_golang_machine,
                            prepend_os_path
                            )

logger = logging.getLogger("kubernator.terragrunt")
proc_logger = logger.getChild("proc")
stdout_logger = StripNL(proc_logger.info)
stderr_logger = StripNL(proc_logger.warning)


class TerragruntPlugin(KubernatorPlugin):
    logger = logger

    _name = "terragrunt"

    def __init__(self):
        self.context = None
        self.tg_file = None
        self.tg_dir = None

    def set_context(self, context):
        self.context = context

    def stanza(self):
        return [self.tg_file]

    def register(self, version=None):
        context = self.context
        context.app.register_plugin("terraform")

        if version:
            # Download and use specific version
            tg_url = (f"https://github.com/gruntwork-io/terragrunt/releases/download/v{version}/"
                      f"terragrunt_{get_golang_os()}_{get_golang_machine()}")
            tg_file_cache, _ = context.app.download_remote_file(logger, tg_url, "bin")

            self.tg_dir = tempfile.TemporaryDirectory()
            context.app.register_cleanup(self.tg_dir)

            tg_file = Path(self.tg_dir.name) / "terragrunt"
            copy(tg_file_cache, tg_file)
            os.chmod(tg_file, 0o500)
            prepend_os_path(self.tg_dir.name)
        else:
            # Use current version
            tg_file = which("terragrunt")
            if not tg_file:
                raise RuntimeError("`terragrunt` cannot be found and no version has been specified")

            logger.debug("Found Terragrunt in %r", tg_file)

        self.tg_file = str(tg_file)

        version_out: str = context.app.run_capturing_out([self.tg_file, "-v"], stderr_logger)
        version = version_out.split(" ")[-1][1:].strip()
        context.globals.terragrunt = dict(version=version,
                                          tg_file=self.tg_file,
                                          stanza=self.stanza,
                                          )

        logger.info("Found Terragrunt version %s at %s", version, self.tg_file)

    def handle_init(self):
        context = self.context
        context.terragrunt = dict(default_includes=Globs(["*.hcl"], True),
                                  default_excludes=Globs([".*"], True),
                                  )

    def handle_before_dir(self, cwd: Path):
        context = self.context
        context.terragrunt.default_includes = Globs(context.terragrunt.default_includes)
        context.terragrunt.default_excludes = Globs(context.terragrunt.default_excludes)
        context.terragrunt.includes = Globs(context.terragrunt.default_includes)
        context.terragrunt.excludes = Globs(context.terragrunt.default_excludes)

    def handle_after_dir(self, cwd: Path):
        context = self.context
        tg = context.terragrunt

        tg_detected = False
        for f in scan_dir(logger, cwd, lambda d: d.is_file(), tg.excludes, tg.includes):
            p = cwd / f.name
            display_p = context.app.display_path(p)
            logger.debug("Detected Terragrunt file in %s", display_p)
            tg_detected = True

        if not tg_detected:
            return

        context.app.run(self.stanza() + ["run-all", "init", "-reconfigure", "-input=false", "-upgrade=false"],
                        stdout_logger, stderr_logger, cwd=cwd).wait()

        output_json = context.app.run_capturing_out(self.stanza() + ["run-all", "output", "-json"],
                                                    stderr_logger, cwd=cwd)

        json_decoder = json.JSONDecoder()
        while output_json:
            output, index = json_decoder.raw_decode(output_json)
            if not index:
                raise RuntimeError("failed to parse JSON output by Terragrunt")

            if "tf" not in context:
                context.tf = {}

            for k, v in output.items():
                context.tf[k] = v["value"]

            output_json = output_json[index:].lstrip()

    def __repr__(self):
        return "Terragrunt Plugin"



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Express Systems USA, Inc
#   Copyright 2024 Karellen, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

# flake8: noqa

URLLIB_HEADERS_PATCH = ("@@ -2909,20 +2909,15 @@\n esp.\n-get\n headers\n-()\n %0A   \n",
                        "kubernetes/client/exceptions.py", "http_resp.headers", None, None, "urllib_headers_patch")
CUSTOM_OBJECT_PATCH_23 = ("""@@ -2517,32 +2517,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -6004,32 +6004,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -8031,33 +8031,65 @@
  'field_manager'
+,%0A            'field_validation'
 %0A
         %5D%0A      
@@ -11267,32 +11267,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -14268,32 +14268,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -17722,32 +17722,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -19774,33 +19774,65 @@
  'field_manager'
+,%0A            'field_validation'
 %0A
         %5D%0A      
@@ -23568,32 +23568,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -36926,32 +36926,160 @@
 pretty printed.%0A
+        :param str label_selector: A selector to restrict the list of returned objects by their labels. Defaults to everything.%0A
         :param i
@@ -40085,32 +40085,160 @@
 pretty printed.%0A
+        :param str label_selector: A selector to restrict the list of returned objects by their labels. Defaults to everything.%0A
         :param i
@@ -42457,32 +42457,62 @@
       'pretty',%0A
+            'label_selector',%0A
             'gra
@@ -45089,32 +45089,247 @@
 )  # noqa: E501%0A
+        if 'label_selector' in local_var_params and local_var_params%5B'label_selector'%5D is not None:  # noqa: E501%0A            query_params.append(('labelSelector', local_var_params%5B'label_selector'%5D))  # noqa: E501%0A
         if 'grac
@@ -48371,32 +48371,160 @@
 pretty printed.%0A
+        :param str label_selector: A selector to restrict the list of returned objects by their labels. Defaults to everything.%0A
         :param i
@@ -51650,32 +51650,160 @@
 pretty printed.%0A
+        :param str label_selector: A selector to restrict the list of returned objects by their labels. Defaults to everything.%0A
         :param i
@@ -54047,32 +54047,62 @@
       'pretty',%0A
+            'label_selector',%0A
             'gra
@@ -57245,32 +57245,247 @@
 )  # noqa: E501%0A
+        if 'label_selector' in local_var_params and local_var_params%5B'label_selector'%5D is not None:  # noqa: E501%0A            query_params.append(('labelSelector', local_var_params%5B'label_selector'%5D))  # noqa: E501%0A
         if 'grac
@@ -157323,32 +157323,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -161012,32 +161012,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -163237,32 +163237,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -166808,32 +166808,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -167725,32 +167725,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -170267,32 +170267,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -173960,32 +173960,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -176185,32 +176185,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -179792,32 +179792,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -180768,32 +180768,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -183320,32 +183320,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -187018,32 +187018,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -189243,32 +189243,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -192856,32 +192856,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -193832,32 +193832,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -196491,32 +196491,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -200302,32 +200302,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -202552,32 +202552,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -206683,32 +206683,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -207600,32 +207600,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -210271,32 +210271,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -214084,32 +214084,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -216334,32 +216334,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -220507,32 +220507,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -221483,32 +221483,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -221530,32 +221530,64 @@
 merge-patch+json
+', 'application/apply-patch+yaml
 '%5D)  # noqa: E50
@@ -224196,32 +224196,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -228014,32 +228014,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -230264,32 +230264,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -234444,32 +234444,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -235420,32 +235420,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -235467,32 +235467,64 @@
 merge-patch+json
+', 'application/apply-patch+yaml
 '%5D)  # noqa: E50
@@ -237883,32 +237883,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -241233,32 +241233,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -243258,33 +243258,65 @@
  'field_manager'
+,%0A            'field_validation'
 %0A
         %5D%0A      
@@ -246821,32 +246821,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -249716,32 +249716,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -253057,32 +253057,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -255082,33 +255082,65 @@
  'field_manager'
+,%0A            'field_validation'
 %0A
         %5D%0A      
@@ -258681,32 +258681,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -261645,32 +261645,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -264991,32 +264991,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -267016,33 +267016,65 @@
  'field_manager'
+,%0A            'field_validation'
 %0A
         %5D%0A      
@@ -270621,32 +270621,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -273705,32 +273705,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -277177,32 +277177,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -279227,33 +279227,65 @@
  'field_manager'
+,%0A            'field_validation'
 %0A
         %5D%0A      
@@ -283352,32 +283352,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -286376,32 +286376,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -289837,32 +289837,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -291887,33 +291887,65 @@
  'field_manager'
+,%0A            'field_validation'
 %0A
         %5D%0A      
@@ -296054,32 +296054,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -299147,32 +299147,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -302613,32 +302613,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -304671,17 +304671,49 @@
 manager'
+,%0A            'field_validation'
 %0A

@@ -308837,32 +308837,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
""", "kubernetes/client/api/custom_objects_api.py", "field_validation", 23, 24, "field_validation_patch_23-24")

CUSTOM_OBJECT_PATCH_25 = ("""@@ -2532,32 +2532,1022 @@
 %0A        :param 
+str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A        :param 
 _preload_content
@@ -3587,32 +3587,32 @@
 nse object will%0A
-

@@ -6004,32 +6004,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -8031,32 +8031,64 @@
  'field_manager'
+,%0A            'field_validation'
 %0A        %5D%0A     
@@ -11267,32 +11267,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -14268,32 +14268,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -17722,32 +17722,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -19774,32 +19774,64 @@
  'field_manager'
+,%0A            'field_validation'
 %0A        %5D%0A     
@@ -23568,32 +23568,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -36926,32 +36926,160 @@
 pretty printed.%0A
+        :param str label_selector: A selector to restrict the list of returned objects by their labels. Defaults to everything.%0A
         :param i
@@ -40085,32 +40085,160 @@
 pretty printed.%0A
+        :param str label_selector: A selector to restrict the list of returned objects by their labels. Defaults to everything.%0A
         :param i
@@ -42457,32 +42457,62 @@
       'pretty',%0A
+            'label_selector',%0A
             'gra
@@ -45089,32 +45089,247 @@
 )  # noqa: E501%0A
+        if 'label_selector' in local_var_params and local_var_params%5B'label_selector'%5D is not None:  # noqa: E501%0A            query_params.append(('labelSelector', local_var_params%5B'label_selector'%5D))  # noqa: E501%0A
         if 'grac
@@ -48371,32 +48371,160 @@
 pretty printed.%0A
+        :param str label_selector: A selector to restrict the list of returned objects by their labels. Defaults to everything.%0A
         :param i
@@ -51650,32 +51650,160 @@
 pretty printed.%0A
+        :param str label_selector: A selector to restrict the list of returned objects by their labels. Defaults to everything.%0A
         :param i
@@ -54047,32 +54047,62 @@
       'pretty',%0A
+            'label_selector',%0A
             'gra
@@ -57245,32 +57245,247 @@
 )  # noqa: E501%0A
+        if 'label_selector' in local_var_params and local_var_params%5B'label_selector'%5D is not None:  # noqa: E501%0A            query_params.append(('labelSelector', local_var_params%5B'label_selector'%5D))  # noqa: E501%0A
         if 'grac
@@ -163053,32 +163053,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -166742,32 +166742,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -168967,32 +168967,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -172538,32 +172538,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -173455,32 +173455,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -175997,32 +175997,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -179690,32 +179690,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -181915,32 +181915,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -185522,32 +185522,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -186498,32 +186498,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -189050,32 +189050,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -192748,32 +192748,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -194973,32 +194973,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -198586,32 +198586,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -199562,32 +199562,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -202221,32 +202221,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -206032,32 +206032,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -208282,32 +208282,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -212413,32 +212413,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -213330,32 +213330,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -216001,32 +216001,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -219814,32 +219814,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -222064,32 +222064,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -226237,32 +226237,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -227213,32 +227213,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -227260,32 +227260,64 @@
 merge-patch+json
+', 'application/apply-patch+yaml
 '%5D)  # noqa: E50
@@ -229926,32 +229926,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -233744,32 +233744,1022 @@
 gicMergePatch).%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param b
@@ -235994,32 +235994,64 @@
 field_manager',%0A
+            'field_validation',%0A
             'for
@@ -240174,32 +240174,255 @@
 )  # noqa: E501%0A
+        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501%0A
         if 'forc
@@ -241150,32 +241150,63 @@
   %5B'application/
+json-patch+json', 'application/
 merge-patch+json
@@ -241197,32 +241197,64 @@
 merge-patch+json
+', 'application/apply-patch+yaml
 '%5D)  # noqa: E50
@@ -243613,32 +243613,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -246963,32 +246963,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -248988,32 +248988,64 @@
  'field_manager'
+,%0A            'field_validation'
 %0A        %5D%0A     
@@ -252551,32 +252551,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -255446,32 +255446,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -258787,32 +258787,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -260812,32 +260812,64 @@
  'field_manager'
+,%0A            'field_validation'
 %0A        %5D%0A     
@@ -264411,32 +264411,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -267375,32 +267375,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -270721,32 +270721,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -272746,32 +272746,64 @@
  'field_manager'
+,%0A            'field_validation'
 %0A        %5D%0A     
@@ -276351,32 +276351,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -279435,32 +279435,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -282907,32 +282907,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -284957,32 +284957,64 @@
  'field_manager'
+,%0A            'field_validation'
 %0A        %5D%0A     
@@ -289082,32 +289082,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -292106,32 +292106,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -295567,32 +295567,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -297617,32 +297617,64 @@
  'field_manager'
+,%0A            'field_validation'
 %0A        %5D%0A     
@@ -301784,32 +301784,255 @@
 ))  # noqa: E501
+%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation'%5D))  # noqa: E501
 %0A%0A        header
@@ -304877,32 +304877,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -308343,32 +308343,1022 @@
 icode/#IsPrint.%0A
+        :param str field_validation: fieldValidation instructs the server on how to handle objects in the request (POST/PUT/PATCH) containing unknown or duplicate fields. Valid values are: - Ignore: This will ignore any unknown fields that are silently dropped from the object, and will ignore all but the last duplicate field that the decoder encounters. This is the default behavior prior to v1.23. - Warn: This will send a warning via the standard warning response header for each unknown field that is dropped from the object, and for each duplicate field that is encountered. The request will still succeed if there are no other errors, and will only persist the last of any duplicate fields. This is the default in v1.23+ - Strict: This will fail the request with a BadRequest error if any unknown fields would be dropped from the object, or if any duplicate fields are present. The error returned from the server will contain all unknown and duplicate fields encountered. (optional)%0A
         :param _
@@ -310401,16 +310401,48 @@
 manager'
+,%0A            'field_validation'
 %0A       
@@ -314469,32 +314469,32 @@
 :  # noqa: E501%0A
-
             quer L
@@ -314549,32 +314549,255 @@
 s%5B'field_manager
+'%5D))  # noqa: E501%0A        if 'field_validation' in local_var_params and local_var_params%5B'field_validation'%5D is not None:  # noqa: E501%0A            query_params.append(('fieldValidation', local_var_params%5B'field_validation
 '%5D))  # noqa: E5
""", "kubernetes/client/api/custom_objects_api.py", "field_validation", 25, None, "field_validation_patch_25+")
